#    Copyright (c) 2018 AT&T Corporation.
#
#    All Rights Reserved.
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.engine.clients.os.neutron import neutron_constraints as nc

CLIENT_NAME = 'neutron'


class TapServiceConstraint(nc.NeutronExtConstraint):
    resource_name = 'tap_service'
    extension = 'taas'


class TapFlowConstraint(nc.NeutronExtConstraint):
    resource_name = 'tap_flow'
    extension = 'taas'


class TaaSProviderConstraint(nc.ProviderConstraint):
    service_type = 'TAPASASERVICE'
