#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.engine.clients import client_plugin
from oslo_log import log as logging
from vitrageclient import client as vitrage_client

LOG = logging.getLogger(__name__)

CLIENT_NAME = 'vitrage'


class VitrageClientPlugin(client_plugin.ClientPlugin):

    exceptions_module = None

    service_types = [RCA] = ['rca']

    def _create(self):
        return vitrage_client.Client('1', self.context.keystone_session)
